<?php

/**
 * @file
 * Drupal stream wrapper implementation for Brandfolder.
 *
 * Implements DrupalStreamWrapperInterface to provide a Brandfolder wrapper
 * using the "bf://" scheme.
 */

/**
 * The stream wrapper class.
 *
 * @todo: Copy generic methods from MediaReadOnlyStreamWrapper and remove dependency.
 */
class BrandfolderStreamWrapper extends MediaReadOnlyStreamWrapper {

  /**
   * Filemime mapping.
   *
   * @var array
   *   Default map for determining filemime types.
   */
  protected static $mimeTypeMapping = NULL;

  /**
   * {@inheritdoc}
   *
   * @todo: The following code doesn't seem to be necessary, but keep it on hand for now.
   * @code
   *   $this->context = stream_context_get_default();
   *   stream_context_set_option($this->context, 'bf', '<option>', '<value>');
   * @endcode
   *
   * @todo: Consider performing common actions here (like loading a Brandfolder API client instance?) and storing results statically, a la:
   * @code
   *   $settings = &drupal_static('BrandfolderStreamWrapper_constructed_settings');
   *   if (!$settings) {
   *     $settings = [];
   *     $settings['client'] = brandfolder_api();
   *     // ...
   *   }
   * @endcode
   */
  public function __construct() {
    $this->base_url = 'https://cdn.brandfolder.io';
  }

  /**
   * Support for stat().
   *
   * @see parent::url_stat()
   */
  // @codingStandardsIgnoreStart
  public function url_stat($url, $flags) {
    // @codingStandardsIgnoreEnd
    // Load file size from DB before proceeding.
    $this->loadFileData($url);

    return $this->stream_stat();
  }

  /**
   * {@inheritdoc}
   */
  public static function getMimeType($uri, $mapping = NULL) {

    // Load the default mime type map.
    if (!isset(self::$mimeTypeMapping)) {
      include_once DRUPAL_ROOT . '/includes/file.mimetypes.inc';
      self::$mimeTypeMapping = file_mimetype_mapping();
    }

    // If a mapping wasn't specified, use the default map.
    if ($mapping == NULL) {
      $mapping = self::$mimeTypeMapping;
    }

    $extension = '';
    $file_parts = explode('.', basename($uri));

    // Remove the first part: a full filename should not match an extension.
    array_shift($file_parts);

    // Iterate over the file parts, trying to find a match.
    // For my.awesome.image.jpeg, we try:
    // - jpeg
    // - image.jpeg
    // - awesome.image.jpeg.
    while ($additional_part = array_pop($file_parts)) {
      $extension = strtolower($additional_part . ($extension ? '.' . $extension : ''));
      if (isset($mapping['extensions'][$extension])) {
        return $mapping['mimetypes'][$mapping['extensions'][$extension]];
      }
    }

    // No mime type matches, so return the default.
    return 'application/octet-stream';
  }

  /**
   * Override parent implementation of setUri().
   */
  public function setUri($uri) {
    $this->uri = $uri;
  }

  /**
   * Support for fopen(), file_get_contents(), file_put_contents() etc.
   *
   * @param string $url
   *   A string containing the path to the file to open.
   * @param string $mode
   *   The file mode ("r", "wb" etc.).
   * @param bitmask $options
   *   A bit mask of STREAM_USE_PATH and STREAM_REPORT_ERRORS.
   * @param string &$opened_url
   *   A string containing the path actually opened.
   *
   * @return bool
   *   TRUE if file was opened successfully.
   *
   * @see parent::stream_open()
   *
   * @todo: Test.
   */
  // @codingStandardsIgnoreStart
  public function stream_open($url, $mode, $options, &$opened_url) {
    // @codingStandardsIgnoreEnd

    // We only handle Read-Only mode by default.
    if ($mode != 'r' && $mode != 'rb') {
      return FALSE;
    }

    if ($options & STREAM_USE_PATH) {
      $opened_url = $url;
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getExternalUrl() {
    // The current approach is to store most of the URL information right in the
    // URI, so we do not need to perform any additional lookups against the
    // BF API or Drupal DB. Thus, a typical URI looks something like
    // "bf://SH123456/as/abc123-echvmo-7qf0za/my_image.jpg."
    // This is not quite as slick or readable as
    // "bf://abc123-echvmo-7qf0za/my_image.jpg"
    // or "bf://abc123-echvmo-7qf0za," but it should be more
    // performant and allow for straightforward management of (a) assets from
    // multiple Brandfolders and/or (b) files corresponding to different
    // attachments on the same BF asset if we choose to support that in the
    // future.
    // The effective limit on BF asset filenames (including extension) is
    // therefore 217 characters.
    $url_options = [
      'absolute' => TRUE,
    ];

    $scheme_prefix = 'bf://';
    $uri_sans_scheme = substr($this->getUri(), strlen($scheme_prefix));
    $query_params = [];

    // Handle image styles.
    $image_style = NULL;
    if (preg_match("/^styles\/([^\/]+)\/bf\/(.*)$/", $uri_sans_scheme, $matches)) {
      $image_style = $matches[1];
      // Remove the style portion of the URI.
      $uri_sans_scheme = $matches[2];
      $all_image_styles = image_styles();
      if (isset($all_image_styles[$image_style])) {
        // Apply all effects from the given image style. Our image toolkit will
        // handle compatible effects and add corresponding Smart CDN URL
        // transformation params to the image object.
        // @todo: Test scenarios with stacked effects; try to provide more robust pass-through support for non BF images.
        $style = $all_image_styles[$image_style];
        if ($image = image_load("bf://$uri_sans_scheme")) {
          foreach ($style['effects'] as $effect) {
            image_effect_apply($image, $effect);
          }
          if (!empty($image->brandfolder['url_params'])) {
            // Safeguard against illegal values for crop zone and final image
            // size. These can occur when the image metadata is missing, etc.
            // The assumption here is that it is better to fall back to a
            // non-transformed image than to nothing.
            if ($image->brandfolder['url_params']['crop_width'] == 0 || $image->brandfolder['url_params']['crop_height'] == 0) {
              unset($image->brandfolder['url_params']['crop_width'], $image->brandfolder['url_params']['crop_height']);
            }
            if ($image->brandfolder['url_params']['width'] == 0 || $image->brandfolder['url_params']['height'] == 0) {
              unset($image->brandfolder['url_params']['width'], $image->brandfolder['url_params']['height']);
            }
            $query_params = array_merge($query_params, $image->brandfolder['url_params']);
          }
        }
      }
    }

    // Lastly, convert the file format/extension to match the globally
    // configured preference if applicable. It's important to do this as late
    // as possible so other modules can accurately map the URI to a managed
    // file.
    if ($default_format = variable_get('brandfolder_default_cdn_file_format', FALSE)) {
      $uri_sans_scheme = preg_replace('/\.\w+$/', ".$default_format", $uri_sans_scheme);
    }

    $url = "{$this->base_url}/{$uri_sans_scheme}";
    // Remove any query params from the original URL and add them to the query
    // params array.
    $url_components = parse_url($url);
    if (!empty($url_components['query'])) {
      $original_query = $url_components['query'];
      $url = str_replace("?$original_query", '', $url);
      $original_query_pairs = explode('&', $original_query);
      foreach ($original_query_pairs as $pair) {
        $key_value = explode('=', $pair);
        if (count($key_value) == 2) {
          [$key, $value] = $key_value;
          $query_params[$key] = $value;
        }
      }
    }
    $url_options['query'] = $query_params;

    drupal_alter('brandfolder_file_external_url', $url, $url_options, $image_style);

    return url($url, $url_options);
  }

  /**
   * {@inheritdoc}
   */
  protected function loadFileData($uri) {
    // @todo: static/cache; data other than file size; etc.
    $result = db_select('brandfolder_file', 'bf')
      ->fields('bf', ['filesize'])
      ->condition('uri', $uri)
      ->execute();

    if ($result->rowCount() > 0) {
      $row = $result->fetch();
      if (!empty($row->filesize)) {
        // Set the appropriate items in the _stat array, which is used to
        // deliver data in response to file-system-esque requests.
        $this->_stat[7] = $this->_stat['size'] = $row->filesize;
      }
    }
  }

  /**
   * DrupalStreamWrapperInterface Implementations.
   */

  /**
   * Gets the path that the wrapper is responsible for.
   *
   * This function isn't part of DrupalStreamWrapperInterface, but the rest
   * of Drupal calls it as if it were, so we need to define it.
   *
   * @return string
   *   The empty string. Since this is a remote stream wrapper,
   *   it has no directory path.
   */
  public function getDirectoryPath() {

    return '';
  }

}
