<?php

/**
 * @file
 * Brandfolder image toolkit.
 *
 * Allow Drupal to handle images that may live in Brandfolder, and, when that
 * is not the case, cleanly fall back to a traditional image toolkit.
 */

/**
 * Settings for the Brandfolder toolkit.
 */
function image_brandfolder_settings() {
  if (brandfolder_api()) {
    $form['status'] = array(
      '#markup' => t('The Brandfolder image toolkit is working properly.'),
    );

    $toolkits = [];
    $toolkit_modules = module_implements('image_toolkits');
    foreach ($toolkit_modules as $module) {
      if ($module == 'brandfolder') {
        continue;
      }
      $module_toolkits = module_invoke($module, 'image_toolkits');
      foreach ($module_toolkits as $toolkit_id => $toolkit_info) {
        if ($toolkit_info['available']) {
          $toolkits[$toolkit_id] = $toolkit_info['title'];
        }
      }
    }

    $form['brandfolder_image_toolkit_fallback'] = array(
      '#type' => 'select',
      '#title' => t('Standard Toolkit'),
      '#description' => t('The image toolkit to use for non-Brandfolder images. If you want to configure the settings for this toolkit here, you can temporarily make it the primary toolkit, choose your configuration, then switch back to Brandfolder as the primary toolkit.'),
      '#options' => $toolkits,
      '#default_value' => variable_get('brandfolder_image_toolkit_fallback', 'gd'),
    );

    return $form;
  }
  else {
    form_set_error('image_toolkit', t('The Brandfolder image toolkit depends upon the Brandfolder PHP API library, which does not seem to be accessible at the moment.'));

    return FALSE;
  }
}

/**
 * Get details about an image.
 *
 * @param object $image
 *   An image object.
 *
 * @return array|bool
 *   FALSE, if the file could not be found or is not an image. Otherwise, a
 *   keyed array containing information about the image:
 *   - "width": Width, in pixels.
 *   - "height": Height, in pixels.
 *   - "mime_type": MIME type.
 *   - "extension": Commonly used file extension for the image.
 *
 * @see image_get_info()
 */
function image_brandfolder_get_info(object $image) {
  $details = FALSE;

  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('get_info', $image);
    }
  }

  $uri_base = preg_replace('/\.\w+(\?.*)?$/', '', $image->source);

  $result = db_select('brandfolder_file', 'bf')
    ->fields('bf', ['width', 'height', 'mime_type'])
    ->condition('uri', "^$uri_base\.[^\.]+$", 'REGEXP')
    ->execute();

  if ($result->rowCount() > 0) {
    $row = $result->fetch();
    if (!empty($row->width)) {
      $details['width'] = $row->width;
    }
    if (!empty($row->height)) {
      $details['height'] = $row->height;
    }
    if (!empty($row->mime_type)) {
      $details['mime_type'] = $row->mime_type;
      $extension_mapping = [
        'image/bmp' => 'bmp',
        'image/gif' => 'gif',
        'image/vnd.microsoft.icon' => 'ico',
        'image/jpeg' => 'jpg',
        'image/png' => 'png',
        'image/svg+xml' => 'svg',
        'image/tiff' => 'tiff',
        'image/webp' => 'webp',
      ];
      if (isset($extension_mapping[$row->mime_type])) {
        $details['extension'] = $extension_mapping[$row->mime_type];
      }
    }
  }

  return $details;
}

/**
 * Basic support for image_load operation.
 *
 * @param object $image
 *   An object representing the image to be loaded.
 *
 * @return bool
 *   Indication of operation's success.
 */
function image_brandfolder_load(object $image) {
  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('load', $image);
    }
  }

  // @todo: probably should set $image->resource to something...
  return TRUE;
}

/**
 * Prepare an image to be scaled, Brandfolder-style.
 *
 * Prepare an image to be scaled to the specified size by Brandfolder's
 * Smart CDN URL transformations (pass through to default image toolkit for
 * non-BF images).
 *
 * @param object $image
 *   An image object. The $image->resource (TBD), $image->info['width'], and
 *   $image->info['height'] values will be modified by this call.
 * @param int $width
 *   The new width of the resized image, in pixels.
 * @param int $height
 *   The new height of the resized image, in pixels.
 *
 * @return bool
 *   TRUE or FALSE, based on success.
 *
 * @see image_resize()
 */
function image_brandfolder_resize(object $image, $width, $height) {

  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('resize', $image, [$width, $height]);
    }
  }

  if (!isset($image->brandfolder)) {
    $image->brandfolder = [
      'original_width' => $image->info['width'],
      'original_height' => $image->info['height'],
      'effect_trace' => [],
      'url_params' => [],
    ];
  }
  $image->brandfolder['effect_trace'][] = 'resize';

  $image->brandfolder['url_params']['width'] = $image->info['width'] = $width;
  $image->brandfolder['url_params']['height'] = $image->info['height'] = $height;

  return TRUE;
}

/**
 * Rotate an image the given number of degrees. Not supported by Brandfolder.
 */
function image_brandfolder_rotate(stdClass $image, $degrees, $background = NULL) {

  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('rotate', $image, [$degrees, $background]);
    }
  }

  return TRUE;
}

/**
 * Prepare an image to be cropped, Brandfolder-style.
 *
 * Prepare an image to be cropped according to the specified
 * coordinates/dimensions by Brandfolder's Smart CDN URL transformations
 * (pass through to default image toolkit for non-BF images).
 *
 * @param object $image
 *   An image object. The $image->resource (TBD), $image->info['width'], and
 *   $image->info['height'] values will be modified by this call.
 * @param int $x
 *   The starting x offset at which to start the crop, in pixels.
 * @param int $y
 *   The starting y offset at which to start the crop, in pixels.
 * @param int $width
 *   The width of the cropped area, in pixels.
 * @param int $height
 *   The height of the cropped area, in pixels.
 *
 * @return bool
 *   TRUE or FALSE, based on success.
 *
 * @see image_crop()
 */
function image_brandfolder_crop(object $image, $x, $y, $width, $height) {

  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('crop', $image, [$x, $y, $width, $height]);
    }
  }

  // Maintain data that will (a) help us determine how to coordinate multiple
  // effects, and (b) allow us to generate the necessary Brandfolder CDN URL
  // after all effects have been applied.
  if (!isset($image->brandfolder)) {
    $image->brandfolder = [
      'original_width' => $image->info['width'],
      'original_height' => $image->info['height'],
      'effect_trace' => [],
      'url_params' => [],
    ];
  }
  $image->brandfolder['effect_trace'][] = 'crop';

  $crop_width = $width;
  $crop_height = $height;

  // Brandfolder crops are applied before scaling. However, some Drupal image
  // style effects ask for scaling to occur before cropping.
  if (in_array('resize', $image->brandfolder['effect_trace'])) {
    // Multiply values by an appropriate scale so that cropping is performed
    // with respect to the original image. I.e. pretend we're reversing the
    // resizing that occurred previously, then calculate crop. The resizing
    // that occurred previously will still be applied ultimately because it
    // causes the $image->brandfolder['url_params'] width and height values to
    // be populated.
    // @todo: Support effect chains that include multiple scale and/or crop effects?
    $scale = max($image->brandfolder['original_width'] / $image->info['width'], $image->brandfolder['original_height'] / $image->info['height']);
    $crop_width *= $scale;
    $crop_height *= $scale;
    $x *= $scale;
    $y *= $scale;
  }

  $image_width = $image->brandfolder['original_width'];
  $image_height = $image->brandfolder['original_height'];
  // Brandfolder does not support upscaling images (and specifying crop offset
  // params for undersized images actually causes the resulting image to include
  // undesirable white padding). If the source image is too small, try to crop
  // it to obtain the largest possible portion of the image that respects the
  // given aspect ratio.
  if ($crop_width > $image_width || $crop_height > $image_height) {
    $aspect_ratio = $crop_width / $crop_height;
    if ($image_width / $aspect_ratio <= $image_height) {
      // Crop the image to remove the excess height.
      $new_height = (int) ($image_width / $aspect_ratio);
      $new_width = (int) $image_width;
    }
    elseif ($image_height * $aspect_ratio <= $image_width) {
      // Crop the image to remove the excess width.
      $new_width = (int) ($image_height * $aspect_ratio);
      $new_height = (int) $image_height;
    }
    if (isset($new_width, $new_height)) {
      $image->brandfolder['url_params']['crop_width'] = $new_width;
      $image->brandfolder['url_params']['crop_height'] = $new_height;
      // Ensure that the actual specified width and height are used for final
      // scaling of the cropped image.
      // @todo: Test with different image effect combos.
      $image->brandfolder['url_params']['width'] = $new_width;
      $image->brandfolder['url_params']['height'] = $new_height;

      $image->info['width'] = $new_width;
      $image->info['height'] = $new_height;
    }
  }
  else {
    // The given $x and $y coordinates represent the upper left corner of the
    // crop zone, measured from the upper left corner of the image.
    // Brandfolder crop zones are defined in relation to the center of the image
    // by default, but we can change that to match Drupal's approach simply by
    // specifying a "crop gravity" of "north_west".
    $image->brandfolder['url_params']['crop_gravity'] = 'north_west';
    $image->brandfolder['url_params']['crop_width'] = (int) $crop_width;
    $image->brandfolder['url_params']['crop_height'] = (int) $crop_height;
    $image->brandfolder['url_params']['crop_x_offset'] = (int) $x;
    $image->brandfolder['url_params']['crop_y_offset'] = (int) $y;
    // Ensure that the actual specified width and height are used for final
    // scaling of the cropped image.
    // @todo: Test with different image effect combos.
    $image->brandfolder['url_params']['width'] = $width;
    $image->brandfolder['url_params']['height'] = $height;

    $image->info['width'] = $width;
    $image->info['height'] = $height;
  }

  return TRUE;
}

/**
 * Convert an image resource to grayscale. Not supported by Brandfolder.
 *
 * @param object $image
 *   An image object.
 *
 * @return bool
 *   TRUE or FALSE, based on success.
 */
function image_brandfolder_desaturate(object $image) {

  $scheme = file_uri_scheme($image->source);
  if ($scheme !== 'bf') {
    $fallback_toolkit = variable_get('brandfolder_image_toolkit_fallback');
    if ($fallback_toolkit) {
      $image->toolkit = $fallback_toolkit;

      return image_toolkit_invoke('desaturate', $image);
    }
  }

  return TRUE;
}
